/********************************************************************************/
/* Matrix product program with MPI on a virtual ring of processors              */
/* S. Vialle - December 2019                                                    */
/********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>
#include <mpi.h>

#include "main.h"
#include "init.h"
#include "kernel.h"


/*------------------------------------------------------------------------------*/
/* Global variable declarations.                                                */
/*------------------------------------------------------------------------------*/

/* Matrixes: C = AxB                                                            */
double *A_Slice;                               /* A matrix.                    */
double *B_Slice;                               /* B matrix.                    */
double *C_Slice;                               /* C matrix (result matrix).    */

unsigned long SIZE;                           /* Total matrix sid              */
unsigned long LOCAL_SIZE;                     /* Nb of lines on 1 process      */

/* Global variables for the management of the parallelisation.                  */
/* Need to be initialized dynamically: set to dummy values for the moment.      */
int Me = -1;                                   /* Processor rank                */
int NbPE = -1;                                 /* Number of processors          */

/* Global variables for the management of the result and performance printing.  */
unsigned long PrinterPE = 0;         /* Processor hosting the central elt of C.*/
unsigned long PrintedElt_i = 0;      /* Coordinates of the central elt of C in */ 
unsigned long PrintedElt_j = 0;      /* its host processor.                    */

/* Global variables to control computations.                                    */
int NbThreads = -1;
int KernelId = -1; 


/*------------------------------------------------------------------------------*/
/* Parallel computation routine                                                 */
/*------------------------------------------------------------------------------*/
void ComputationAndCirculation()
{
 unsigned long step;       /* Step counter                                     */
 MPI_Status status;         /* Var to store the SUCCESS/ERROR code of the comm. */

 /* Computation-Circulation loop achived by each MPI process.                   */
 for (step = 0; step < NbPE; step++) {
   /* - local computation of a small C square  */
   OneLocalProduct(step);
   /* - local contribution to data circulation */
   MPI_Sendrecv_replace(A_Slice,LOCAL_SIZE*SIZE,MPI_DOUBLE,
                        (Me-1+NbPE)%NbPE,step,(Me+1)%NbPE,step,
                        MPI_COMM_WORLD,&status);
 }
}


/*------------------------------------------------------------------------------*/
/* Toplevel function.                                                           */
/*------------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
 double td1, tf1;                      /* Time measures of the computation loop */
 double d1;                            /* Elapsed times to measure.             */
 double gigaflops;                     /* Program performances to measure.      */

 /* Initialisations ----------------------------------------------------------- */
 MPI_Init(&argc,&argv);                          /* MPI initialisation.         */
 MPI_Comm_size(MPI_COMM_WORLD,&NbPE);            /* Init NbPE global var.       */
 MPI_Comm_rank(MPI_COMM_WORLD,&Me);              /* Init Me global var.         */
 CommandLineParsing(argc,argv);                  /* Cmd line parsing.           */
 ProcessInit();                                  /* Init of the process.        */
 LocalMatrixAllocInit();                         /* Initialization of the data. */
 PrintStartingMsg();                              /* Print comput. configuration */

 /* Matrix product computation ------------------------------------------------ */
 MPI_Barrier(MPI_COMM_WORLD);                    /* To sync. time measures...   */
 td1 = MPI_Wtime();                              /* Begin of comp. time measure.*/
 ComputationAndCirculation();                    /* Parallel Matrix product.    */
 MPI_Barrier(MPI_COMM_WORLD);                    /* To sync. time measures...   */
 tf1 = MPI_Wtime();                              /* End of comp. time measure.  */

 /* Performance computation, results and performance printing ----------------- */
 d1 = tf1 - td1;                                 /* Elapsed comp. time.         */
 gigaflops = (2.0*pow(SIZE,3))/d1*1E-9;          /* Performance achieved.       */
 PrintResultsAndPerf(gigaflops,d1);              /* Results and perf printing   */

 /* Parallel program ending --------------------------------------------------- */
 LocalMatrixCleanup();
 fflush(stdout);
 MPI_Barrier(MPI_COMM_WORLD);
 MPI_Finalize();                                 /* End of the MPI usage.       */
 return(EXIT_SUCCESS);
}

